{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "Fue9xn6CKNvf"
      },
      "source": [
        "# 1. Classes and basic syntax\n",
        "\n",
        "Author: Rasmus Ørsøe"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "-oLjknp6RN72"
      },
      "source": [
        "Classes are the central piece of *object-oriented programming*. They allow encapsulation of functions and data into logical units or *classes*.\n",
        "The main part of a class are its methods (functions of the class) and the attributes (member variables of the class).\n",
        "Note that there are no private variables or functions in Python. In general, you can access every function and every member variable of an object.\n",
        "\n",
        "We will see below how to construct a class and how to create an object from it."
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "1Vp4QuYriRRr"
      },
      "source": [
        "Basic rule: everything in Python is an object! Even if you define a simple number:\n",
        "\n",
        "a = 3\n",
        "\n",
        "You are actually creating an 'int' object. Let's take a closer look at the class syntax."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 1,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "x4Cuc0QmHaEI",
        "outputId": "3e23f205-29bd-46bf-da46-2d209deb893c"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "this is my first class\n"
          ]
        }
      ],
      "source": [
        "class a_descriptive_name():\n",
        "    # A very simple class\n",
        "    print('this is my first class')\n",
        ""
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "NieEpkVjwgJN"
      },
      "source": [
        "In the cell above, the class is defined by writing `class my_class_name():` - which is nearly the same as defining a function, but with the defining difference of exchanging `def` with `class`.\n",
        "\n",
        "When run the cell in which the class is defined, the code inside the class is run. That's why we see the print message in the terminal.\n",
        "\n",
        "Classes are objects; so unlike functions, classes are python objects that you **instantiate** and can interact with later on. You can instatiate the class above by running the cell below."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 2,
      "metadata": {
        "id": "uOPudBc8IKad"
      },
      "outputs": [],
      "source": [
        "my_instantiated_class = a_descriptive_name()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "QuXGCxJYxLsB"
      },
      "source": [
        "When you instantiate a class, you create a copy of the class that you can interact with. Technically, the class is constructed and how the construction is done depends on the `__init__` function. We wrote a very simple class that doesn't have one. Let's try to make a version of the class that has a simple constructor function."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 3,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "nxkcbXfcIlvy",
        "outputId": "4d13678e-577e-418d-dcb6-6afa161583c8"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "this is my first class\n"
          ]
        }
      ],
      "source": [
        "class a_descriptive_name():\n",
        "    # A very simple class\n",
        "    print('this is my first class')\n",
        "\n",
        "    def __init__(self, my_message):\n",
        "      self.message = my_message # this saves the message as a \"member variable\"\n",
        "      print(self.message)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "cjHKdb99xo3B"
      },
      "source": [
        "So again, when we execute the cell above, the code right underneath the class definition is run; so we see the print message - but it also means that our definition of the `__init__` function is run! Therefore, the `__init__` function **exists as a piece of code only inside the class**!\n",
        "\n",
        "Also, as you can see, our simple `__init__` function just saves the message as a \"member variable\" and prints a message that it gets as input. When we ran the cell above, we did not see this message. This is because the `__init__` is only executed when we instantiate the class. Let's try to instantiate the class with a message.\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 4,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "6TJ2boAWIwl_",
        "outputId": "b3ed44bf-1c46-4236-f68c-b899da224c89"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "I am initialized\n"
          ]
        }
      ],
      "source": [
        "my_class = a_descriptive_name(my_message = 'I am initialized')"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "RnNyVs8eykj5"
      },
      "source": [
        "Voila! It works. Lets try to access the original message. This can be done as shown below"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 5,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/",
          "height": 35
        },
        "id": "BGPmts7LL6Dw",
        "outputId": "038b24da-3fe2-4a19-fd2e-3400294e617d"
      },
      "outputs": [
        {
          "output_type": "execute_result",
          "data": {
            "text/plain": [
              "'I am initialized'"
            ],
            "application/vnd.google.colaboratory.intrinsic+json": {
              "type": "string"
            }
          },
          "metadata": {},
          "execution_count": 5
        }
      ],
      "source": [
        "my_class.message"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "GUxCnqpMyrOZ"
      },
      "source": [
        "So now you've seen how to define and instantiate a very simple class. You might correctly point out that all we have achieved now is to make the goal of printing a message to the terminal unneccesarily complicated.\n",
        "\n",
        "Classes is an abstract tool in python that allows you to create programmable objects that you can interact with. Therefore, simple tasks as calculating a number, printing a message etc. are usually most simply solved by using functions. However, more complicated tasks can become more managable if presented as classes instead of functions. Below we'll go through an example of that with animals and boxes!"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "YLKk0pabzWKB"
      },
      "source": [
        "# 2. The Animal Class & \"public\" methods\n",
        "\n",
        "\n",
        "\n",
        "Let's create a python class that represents an animal. It could be any animal - but for our purpose we want the animal to have 4 member variables; a species, an age, a noise and a name. Also, we want the Animal to have a \"public method\" that we can call later on. Let's define it!"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 6,
      "metadata": {
        "id": "IvdiWFUDLdBm"
      },
      "outputs": [],
      "source": [
        "class Animal():\n",
        "    \"\"\"\n",
        "    We now define the 'init' constructor.\n",
        "    This function is automaticalle called, when we're instantiating the class.\n",
        "    Every argument has to be supplied when instantiating.\n",
        "\n",
        "    \"\"\"\n",
        "    def __init__(self, species, age, noise, name):\n",
        "        \"\"\"\n",
        "        The first argument of the constructor (and all instance methods) is always\n",
        "        the instantiated object. With this variable we can access attributes, that are specific\n",
        "        to this respective instance.\n",
        "        \"\"\"\n",
        "        # here we set the 'member variables' of the class\n",
        "        self.species = species\n",
        "        self.age = age\n",
        "        self.noise = noise\n",
        "        self.name = name\n",
        "\n",
        "    def make_noise(self):\n",
        "      # A public function;\n",
        "      #by convention does not have '_' in beginning of name.\n",
        "      # Meant to be called after the class is instantiated.\n",
        "      print(f'{self.name} says: {self.noise}')\n",
        "      return"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "CcGVb1SvzzIZ"
      },
      "source": [
        "In this slightly more complicated class we've defined a `__init__` function tha t accepts species, age, noise and name as inputs, and that saves these as member variables. We've then defined a function called \"make_noise\" that prints the name of the animal and the noise that the animal makes by accessing the assigned member variables. Let's try to instantiate this class for a cat named Garfield!"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 7,
      "metadata": {
        "id": "lyoacbNBMjUb"
      },
      "outputs": [],
      "source": [
        "cat = Animal(species = 'cat', age = 3, noise = 'meow', name = 'Garfield')\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "rJ9aIr-U0lUl"
      },
      "source": [
        "Let's try to access the member variables in `cat`"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 8,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "yctvWmJT0qGE",
        "outputId": "90fafabe-da56-4e5e-d23f-c7c6943a2a90"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Garfield cat 3 meow\n"
          ]
        }
      ],
      "source": [
        "print(cat.name, cat.species, cat.age, cat.noise)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "VgntirQb0uYY"
      },
      "source": [
        "Alright - that looks good. Let's now try to use the public method of our cat:"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 9,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "jTXSs8VoO6lm",
        "outputId": "6a2fd191-da1f-4cb2-8628-2afa9d76a4ee"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Garfield says: meow\n"
          ]
        }
      ],
      "source": [
        "cat.make_noise()"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 10,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "wxvj5M1FQZ0d",
        "outputId": "fd52e7d6-3c8f-40b5-e9f7-90803a0aff99"
      },
      "outputs": [
        {
          "output_type": "execute_result",
          "data": {
            "text/plain": [
              "['__class__',\n",
              " '__delattr__',\n",
              " '__dir__',\n",
              " '__doc__',\n",
              " '__eq__',\n",
              " '__format__',\n",
              " '__ge__',\n",
              " '__getattribute__',\n",
              " '__gt__',\n",
              " '__hash__',\n",
              " '__init__',\n",
              " '__init_subclass__',\n",
              " '__le__',\n",
              " '__lt__',\n",
              " '__ne__',\n",
              " '__new__',\n",
              " '__reduce__',\n",
              " '__reduce_ex__',\n",
              " '__repr__',\n",
              " '__setattr__',\n",
              " '__sizeof__',\n",
              " '__str__',\n",
              " '__subclasshook__']"
            ]
          },
          "metadata": {},
          "execution_count": 10
        }
      ],
      "source": [
        "dir(object)"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 11,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "nq88EDJCP9jY",
        "outputId": "84d084ea-348a-4d37-80cd-2fcaccc8d8d0"
      },
      "outputs": [
        {
          "output_type": "execute_result",
          "data": {
            "text/plain": [
              "['__class__',\n",
              " '__delattr__',\n",
              " '__dict__',\n",
              " '__dir__',\n",
              " '__doc__',\n",
              " '__eq__',\n",
              " '__format__',\n",
              " '__ge__',\n",
              " '__getattribute__',\n",
              " '__gt__',\n",
              " '__hash__',\n",
              " '__init__',\n",
              " '__init_subclass__',\n",
              " '__le__',\n",
              " '__lt__',\n",
              " '__module__',\n",
              " '__ne__',\n",
              " '__new__',\n",
              " '__reduce__',\n",
              " '__reduce_ex__',\n",
              " '__repr__',\n",
              " '__setattr__',\n",
              " '__sizeof__',\n",
              " '__str__',\n",
              " '__subclasshook__',\n",
              " '__weakref__',\n",
              " 'age',\n",
              " 'make_noise',\n",
              " 'name',\n",
              " 'noise',\n",
              " 'species']"
            ]
          },
          "metadata": {},
          "execution_count": 11
        }
      ],
      "source": [
        "dir(cat)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "IBhFZsby0zu0"
      },
      "source": [
        "So as you can see; we have successfully represented a cat as a python class, and we are able to interact with this object in our code after it's been instantiated. Now what? Private methods!\n",
        "\n",
        "Technically speaking, there are no private methods in python. All functions defined inside a class can be accessed \"outside\" the class. However, it is good practice to designate if a function is supposed to be treated as a public method or not by giving the private methods a `_` in or `__` in the beginning of their name.\n",
        "\n",
        "# 2.1 \"private\" methods\n",
        "\n",
        "\n",
        "\n",
        "So let's now introduce a new member variable; the mood of the animal. We want to write a private method that uses this member variable to set some internal logic. In our case, we want to change the noise that the animal makes if the mood happens to be \"angry\". See below.\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 12,
      "metadata": {
        "id": "RmgUi3P1NLoH"
      },
      "outputs": [],
      "source": [
        "class Animal():\n",
        "    \"\"\"\n",
        "    We now define the 'init' constructor.\n",
        "    This function is automaticalle called, when we're instantiating the class.\n",
        "    Every argument has to be supplied when instantiating.\n",
        "\n",
        "    \"\"\"\n",
        "    def __init__(self, species, age, noise, name, mood = 'happy'):\n",
        "        \"\"\"\n",
        "        The first argument of the constructor (and all instance methods) is always\n",
        "        the instantiated object. With this variable we can access attributes, that are specific\n",
        "        to this respective instance.\n",
        "        \"\"\"\n",
        "        # here we set the 'member variables' of the class\n",
        "        self.species = species\n",
        "        self.age = age\n",
        "        self.noise = noise\n",
        "        self.mood = mood\n",
        "        self.name = name\n",
        "        # Here we call the private function to run it's logic of setting the mood of the animal\n",
        "        self._set_mood()\n",
        "\n",
        "    def make_noise(self):\n",
        "      # A public function;\n",
        "      # by convention does not have '_' in beginning of name.\n",
        "      # Meant to be called after the class is instantiated.\n",
        "        print(f'{self.name} says: {self.noise}')\n",
        "        return\n",
        "\n",
        "    def _set_mood(self):\n",
        "        # A private function;\n",
        "        # By convention starts with '_' .\n",
        "        # Meant to be called by the class itself, and not by the user.\n",
        "\n",
        "        # Here we change the member variable 'self.noise' according to the mood of the animal.\n",
        "        if self.mood == 'angry':\n",
        "            self.noise = self.noise.upper() + '!!'\n",
        "        else:\n",
        "            pass\n",
        "\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "CdWSo68x1pKF"
      },
      "source": [
        "Here the `_set_mood(self)` function change self.noise member variable if the mood is \"angry\" from it's original format to all uppercase with !! . Notice that our new `__init__` function calls this private `set_mood()` function when the class is instantiated - so the logic is trigged when the class is instantiated.\n",
        "\n",
        "Let's check that it works.\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 13,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "91oTKfRANm_S",
        "outputId": "9c11e62c-6141-4c54-eaea-dc382cf0eaf1"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Angry Garfield says: MEOW!!\n"
          ]
        }
      ],
      "source": [
        "angry_cat = Animal(species = 'cat', age = 3, noise = 'meow', name = 'Angry Garfield',mood = 'angry',)\n",
        "angry_cat.make_noise()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "Aho6qrEI2Rmw"
      },
      "source": [
        "Looks good! So now you've seen that we can represent animals as python classes, and that we can use public methods to interact with the animal. You've also seen an example of how internal class logic can be trigged by using \"private\" methods.\n",
        "\n",
        "# 2.2 BoxOfAnimals - Classes interacting with Classes\n",
        "\n",
        "\n",
        "\n",
        "Let's now try to increase the abstraction by defining a new python class that we can put our animals into; the box of animals!"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 14,
      "metadata": {
        "id": "ctxFtX1YTj8N"
      },
      "outputs": [],
      "source": [
        "class BoxOfAnimals():\n",
        "    \"\"\"\n",
        "    We now define the 'init' constructor.\n",
        "    This function is automaticalle called, when we're instantiating the class.\n",
        "    Every argument has to be supplied when instantiating.\n",
        "\n",
        "    \"\"\"\n",
        "    def __init__(self, animal):\n",
        "        \"\"\"\n",
        "        The first argument of the constructor (and all instance methods) is always\n",
        "        the instantiated object. With this variable we can access attributes, that are specific\n",
        "        to this respective instance.\n",
        "        \"\"\"\n",
        "        # member variable\n",
        "        self.inventory = {animal.species: [animal]}\n",
        "\n",
        "    def print_inventory(self):\n",
        "        # Prints the animals in the box\n",
        "        manifest = 'Box contains: \\n'\n",
        "        for species in self.inventory.keys():\n",
        "            manifest = manifest + f'{species}s : \\n'\n",
        "            for animal in self.inventory[species]:\n",
        "                manifest = manifest + f'        {animal.name} \\n'\n",
        "        print(manifest)\n",
        "\n",
        ""
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "nCzvkPX72mEI"
      },
      "source": [
        "Here we define `BoxOfAnimals` as a python class that accepts an Animal class as input to its `__init__` function. The `__init__` creates a member variable called \"inventory\" that is a python dictionary that contains the animals that are in the box. We \"file\" the Animal classes according to their species. Notice that we specify the keys of this dictionary by accessing the member variables of the Animal class that the `BoxOfAnimals` are instatiated with!\n",
        "\n",
        "Secondly, we define a \"public\" method called `print_inventory` that prints all the animals that are in the `inventory` member variable.\n",
        "\n",
        "Let's check that it works by putting Garfield in the box."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 15,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "dTiUOJhHUg_c",
        "outputId": "df15b8d7-df77-412c-bc86-149a87037164"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Box contains: \n",
            "cats : \n",
            "        Garfield \n",
            "\n"
          ]
        }
      ],
      "source": [
        "box = BoxOfAnimals(cat)\n",
        "box.print_inventory()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "6JmmV6gG3XJm"
      },
      "source": [
        "Alright - now we've instantiated a BoxOfAnimals with Garfield in it!\n",
        "\n",
        "Let's now create a new \"public\" method that allows us to put another animal in the already instantiated box."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 16,
      "metadata": {
        "id": "G7GSwUMQVPIn"
      },
      "outputs": [],
      "source": [
        "class BoxOfAnimals():\n",
        "    \"\"\"\n",
        "    We now define the 'init' constructor.\n",
        "    This function is automaticalle called, when we're instantiating the class.\n",
        "    Every argument has to be supplied when instantiating.\n",
        "\n",
        "    \"\"\"\n",
        "    def __init__(self, animal):\n",
        "        \"\"\"\n",
        "        The first argument of the constructor (and all instance methods) is always\n",
        "        the instantiated object. With this variable we can access attributes, that are specific\n",
        "        to this respective instance.\n",
        "        \"\"\"\n",
        "        # member variable\n",
        "        self.inventory = {animal.species: [animal]}\n",
        "\n",
        "    def print_inventory(self):\n",
        "        # Prints the animals in the box\n",
        "        manifest = 'Box contains: \\n'\n",
        "        for species in self.inventory.keys():\n",
        "            manifest = manifest + f'{species}s : \\n'\n",
        "            for animal in self.inventory[species]:\n",
        "                manifest = manifest + f'        {animal.name} \\n'\n",
        "        print(manifest)\n",
        "\n",
        "    def put(self, animal):\n",
        "        # Puts an animal in the box.\n",
        "        if animal.species in self.inventory.keys():\n",
        "            self.inventory[animal.species].append(animal)\n",
        "        else:\n",
        "            self.inventory.update({animal.species: [animal]})\n",
        ""
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "HdPhmjSM3mc8"
      },
      "source": [
        "So this new \"public\" method `put` takes as argument an `Animal` class and checks if there already is an animal with the same species in the box. If so, it puts it in the same category of species, and if not, it creates a new category.\n",
        "\n",
        "Let's re-instatiate this new version of our box with Garfield in it, and then let's try to add angry garfield afterwards using the new method."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 17,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "cmhbeivbVwyS",
        "outputId": "46f8f1de-52ae-42d6-e9c4-e74296e450f5"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Box contains: \n",
            "cats : \n",
            "        Garfield \n",
            "\n"
          ]
        }
      ],
      "source": [
        "box= BoxOfAnimals(cat)\n",
        "box.print_inventory()"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 18,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "P8O1Bb1_Vzzx",
        "outputId": "9e7bd34c-6094-4997-e036-8525e4251023"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Box contains: \n",
            "cats : \n",
            "        Garfield \n",
            "        Angry Garfield \n",
            "\n"
          ]
        }
      ],
      "source": [
        "box.put(angry_cat)\n",
        "box.print_inventory()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "I29rclo74Gxg"
      },
      "source": [
        "Alright! Now we have two cats in the box. Let's try to add a dog too!"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 19,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "bMfkVda1ZV4f",
        "outputId": "bbcfda12-a62e-4644-db55-4793e842caae"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Box contains: \n",
            "cats : \n",
            "        Garfield \n",
            "        Angry Garfield \n",
            "dogs : \n",
            "        Snoopy \n",
            "\n"
          ]
        }
      ],
      "source": [
        "dog = Animal(species = 'dog', age = 2, noise = 'wuff', name = 'Snoopy')\n",
        "box.put(dog)\n",
        "box.print_inventory()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "e-xQNTCk4L6t"
      },
      "source": [
        "Cool! We could put **any** kind of instantiated `Animal` class to the box. But what if we had two boxes, could we add those?\n",
        "\n",
        "# 2.3 `class + class = `?\n",
        "\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 20,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/",
          "height": 179
        },
        "id": "lgg-UbYufCNf",
        "outputId": "3832766b-30ad-403a-e462-f881f7e3bccc"
      },
      "outputs": [
        {
          "output_type": "error",
          "ename": "TypeError",
          "evalue": "ignored",
          "traceback": [
            "\u001b[0;31m---------------------------------------------------------------------------\u001b[0m",
            "\u001b[0;31mTypeError\u001b[0m                                 Traceback (most recent call last)",
            "\u001b[0;32m<ipython-input-20-951a73f27d85>\u001b[0m in \u001b[0;36m<cell line: 1>\u001b[0;34m()\u001b[0m\n\u001b[0;32m----> 1\u001b[0;31m \u001b[0mbox\u001b[0m \u001b[0;34m+\u001b[0m \u001b[0mbox\u001b[0m\u001b[0;34m\u001b[0m\u001b[0;34m\u001b[0m\u001b[0m\n\u001b[0m",
            "\u001b[0;31mTypeError\u001b[0m: unsupported operand type(s) for +: 'BoxOfAnimals' and 'BoxOfAnimals'"
          ]
        }
      ],
      "source": [
        "box + box"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "RKnOJY0h4YJn"
      },
      "source": [
        "This error tells us that python does not know how to add these two objects. We can tell python how it's supposed to use the `+` operator on our `BoxOfAnimals` callses by defining a `__add__` method. (You'll do something similar in the exercises!)\n",
        "\n",
        "\n",
        "Let's try to define this `__add__` in our box class. We want python to understand `box1 + box2` as\n",
        "\n",
        "*Please take all animals in box1 and box2 and put them in a new box* . See below\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 21,
      "metadata": {
        "id": "MTnS5p0jr71_"
      },
      "outputs": [],
      "source": [
        "class BoxOfAnimals():\n",
        "    \"\"\"\n",
        "    We now define the 'init' constructor.\n",
        "    This function is automaticalle called, when we're instantiating the class.\n",
        "    Every argument has to be supplied when instantiating.\n",
        "\n",
        "    \"\"\"\n",
        "    def __init__(self, animal):\n",
        "        \"\"\"\n",
        "        The first argument of the constructor (and all instance methods) is always\n",
        "        the instantiated object. With this variable we can access attributes, that are specific\n",
        "        to this respective instance.\n",
        "        \"\"\"\n",
        "        # member variable\n",
        "        self.inventory = {animal.species: [animal]}\n",
        "\n",
        "    def print_inventory(self):\n",
        "        # Prints the animals in the box\n",
        "        manifest = 'Box contains: \\n'\n",
        "        for species in self.inventory.keys():\n",
        "            manifest = manifest + f'{species}s : \\n'\n",
        "            for animal in self.inventory[species]:\n",
        "                manifest = manifest + f'        {animal.name} \\n'\n",
        "        print(manifest)\n",
        "\n",
        "    def put(self, animal):\n",
        "        # Puts an animal in the box\n",
        "        if animal.species in self.inventory.keys():\n",
        "            self.inventory[animal.species].append(animal)\n",
        "        else:\n",
        "            self.inventory.update({animal.species: [animal]})\n",
        "\n",
        "    def __add__ (self, other_box):\n",
        "      # Adds the animals in the other box to this box.\n",
        "        is_first = True\n",
        "        for species in other_box.inventory.keys():\n",
        "            for animal in other_box.inventory[species]:\n",
        "                if is_first:\n",
        "                    new_box = BoxOfAnimals(animal)\n",
        "                    new_box.inventory = self.inventory\n",
        "                    new_box.put(animal)\n",
        "                    is_false = False\n",
        "                else:\n",
        "                    new_box.put(animal)\n",
        "        return new_box\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "Ju06PiZk5YmJ"
      },
      "source": [
        "Alright, let's try to now put garfield, angry garfield and snoopy in three boxes seperately and then add the three boxes together."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 22,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "cZ5FHfz9sXIj",
        "outputId": "cfc5d95a-0643-415a-fc85-417bf69bb9a4"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Box contains: \n",
            "cats : \n",
            "        Garfield \n",
            "\n"
          ]
        }
      ],
      "source": [
        "box1= BoxOfAnimals(cat)\n",
        "box1.print_inventory()"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 23,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "RbqcCTYWsdQ1",
        "outputId": "9cf1c735-15e6-4917-dc0c-5904ff15c8d4"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Box contains: \n",
            "cats : \n",
            "        Angry Garfield \n",
            "\n"
          ]
        }
      ],
      "source": [
        "box2= BoxOfAnimals(angry_cat)\n",
        "box2.print_inventory()"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 24,
      "metadata": {
        "id": "UOSXJTcbwGA7"
      },
      "outputs": [],
      "source": [
        "box3 = BoxOfAnimals(dog)"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 25,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "25gdWNA3sji8",
        "outputId": "e8e08b0b-bfbb-4324-fb7d-4f705952f694"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Box contains: \n",
            "cats : \n",
            "        Garfield \n",
            "        Angry Garfield \n",
            "dogs : \n",
            "        Snoopy \n",
            "\n"
          ]
        }
      ],
      "source": [
        "new_box = box1 + box2 + box3\n",
        "new_box.print_inventory()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "wMH3Ixyj5hXx"
      },
      "source": [
        "Alright! Perfect - We get a box with all three animals in it. But it would be nice if we could use `len(box)` to tell us how many animals were in the box.."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 26,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/",
          "height": 179
        },
        "id": "lSLPmA895unO",
        "outputId": "a703c285-6a6c-4dbc-b6e1-8ad756eca62d"
      },
      "outputs": [
        {
          "output_type": "error",
          "ename": "TypeError",
          "evalue": "ignored",
          "traceback": [
            "\u001b[0;31m---------------------------------------------------------------------------\u001b[0m",
            "\u001b[0;31mTypeError\u001b[0m                                 Traceback (most recent call last)",
            "\u001b[0;32m<ipython-input-26-4212dd4b6511>\u001b[0m in \u001b[0;36m<cell line: 1>\u001b[0;34m()\u001b[0m\n\u001b[0;32m----> 1\u001b[0;31m \u001b[0mlen\u001b[0m\u001b[0;34m(\u001b[0m\u001b[0mnew_box\u001b[0m\u001b[0;34m)\u001b[0m\u001b[0;34m\u001b[0m\u001b[0;34m\u001b[0m\u001b[0m\n\u001b[0m",
            "\u001b[0;31mTypeError\u001b[0m: object of type 'BoxOfAnimals' has no len()"
          ]
        }
      ],
      "source": [
        "len(new_box)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "ta2jec2i5w0X"
      },
      "source": [
        "But as you see, python doesn't know how to use the `len()` function on our box class. We can define how to count the number of animals in the box for python by defining a `__len__` method in our class. Let's try this.\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 27,
      "metadata": {
        "id": "WKpMQ1cwfGTP"
      },
      "outputs": [],
      "source": [
        "class BoxOfAnimals():\n",
        "    \"\"\"\n",
        "    We now define the 'init' constructor.\n",
        "    This function is automaticalle called, when we're instantiating the class.\n",
        "    Every argument has to be supplied when instantiating.\n",
        "\n",
        "    \"\"\"\n",
        "    def __init__(self, animal):\n",
        "        \"\"\"\n",
        "        The first argument of the constructor (and all instance methods) is always\n",
        "        the instantiated object. With this variable we can access attributes, that are specific\n",
        "        to this respective instance.\n",
        "        \"\"\"\n",
        "        # member variable\n",
        "        self.inventory = {animal.species: [animal]}\n",
        "\n",
        "    def print_inventory(self):\n",
        "        # Prints the animals in the box\n",
        "        manifest = 'Box contains: \\n'\n",
        "        for species in self.inventory.keys():\n",
        "            manifest = manifest + f'{species}s : \\n'\n",
        "            for animal in self.inventory[species]:\n",
        "                manifest = manifest + f'        {animal.name} \\n'\n",
        "        print(manifest)\n",
        "\n",
        "    def put(self, animal):\n",
        "        # Puts an animal in the box\n",
        "        if animal.species in self.inventory.keys():\n",
        "            self.inventory[animal.species].append(animal)\n",
        "        else:\n",
        "            self.inventory.update({animal.species: [animal]})\n",
        "\n",
        "    def __add__ (self, other_box):\n",
        "      # Adds the animals in the other box to this box.\n",
        "        is_first = True\n",
        "        for species in other_box.inventory.keys():\n",
        "            for animal in other_box.inventory[species]:\n",
        "                if is_first:\n",
        "                    new_box = BoxOfAnimals(animal)\n",
        "                    new_box.inventory = self.inventory\n",
        "                    new_box.put(animal)\n",
        "                    is_false = False\n",
        "                else:\n",
        "                    new_box.put(animal)\n",
        "        return self\n",
        "\n",
        "    def __len__ (self):\n",
        "        # Counts the number of animals in the box\n",
        "        n_animals = 0\n",
        "        for species in self.inventory.keys():\n",
        "            for animal in self.inventory[species]:\n",
        "                n_animals +=1\n",
        "        return n_animals"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "HlrjweRQ6BC5"
      },
      "source": [
        "Alright; so our `__len__` function just counts all the animals we have in the member variable `self.inventory`. Let's make a box with garfield in it, see how many animals that correspond to (should be 1) and then lets use the `.put` function to add angry garfield and then count again. (Should be 2)"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 28,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "62NugdOTt8kL",
        "outputId": "dbc4abe3-2fc3-4223-ac30-aa93170525c9"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "1\n",
            "2\n"
          ]
        }
      ],
      "source": [
        "box= BoxOfAnimals(cat)\n",
        "print(len(box))\n",
        "box.put(angry_cat)\n",
        "print(len(box))"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "n0ph1paC6RsM"
      },
      "source": [
        "Perfect! So to summarize, You have now seen:\n",
        "\n",
        "\n",
        "\n",
        "1.   The basic syntax of defining and instantiating classes\n",
        "2.   Handling arguments to the `__init__`\n",
        "3.   Assigning and acessing member variables\n",
        "4.   Manipulating member variables with \"private\" methods\n",
        "5.   Interacting with python classes through \"public\" methods.\n",
        "6.   How to define your classes such that they are compatible with `+` and `len()`\n",
        "\n",
        "In the associated exercises, you'll be tasked with creating a `Vector` Class and implementing vector addition, substraction and dot product using `+` , `-` and `*`. You'll also have to implement a public method `Vector.cross(another_vector)` that calculates the cross product of the vectors!\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "XXcVrCasE5g7"
      },
      "source": [
        "# 3. Elegant Abstraction (Optional)\n",
        "\n",
        "\n",
        "\n",
        "You can work through this part on your own, if you are interested!"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "oKMOCMc7L4FT"
      },
      "source": [
        "## 3.1 Abstract Classes & Decorators\n",
        "\n",
        "In python a decorator is a function which accepts a function as an argument and extends its functionality. Decorator is called with `@` right before defining the function.\n",
        "\n",
        "While creating a class, we can use **property** and **abstractmethod** decorators to create a so-called abstract class. If we also use class inheritance, then we can make our distinction between animals less complicated. In addition, we can use `assert` and `isinstance` to make sanity checks on the inputs to the class. This way, we make sure that any future users of our code uses it correctly."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 29,
      "metadata": {
        "id": "8K9ZR9oUQjYu"
      },
      "outputs": [],
      "source": [
        "from abc import abstractmethod, ABC\n",
        "class Animal(ABC):\n",
        "    \"\"\"\n",
        "    An abstract base class for all animals\n",
        "    \"\"\"\n",
        "    def __init__(self, species: str , age: int, noise: str, name: str, mood: str = 'happy'):\n",
        "        # Here we make checks on the inputs, to make sure that everything is as expected\n",
        "        assert isinstance(species, str), f' \"species\" must be string, got {type(species)}'\n",
        "        assert isinstance(noise, str), f' \"noise\" must be string, got {type(noise)}'\n",
        "        assert isinstance(name, str), f' \"name\" must be string, got {type(name)}'\n",
        "        assert isinstance(mood, str), f' \"mood\" must be string, got {type(mood)}'\n",
        "        assert isinstance(age, int), f' \"age\" must be integer, got {type(age)}'\n",
        "\n",
        "        # Here we call the private function to run it's logic of setting the mood of the animal\n",
        "        self._set_mood()\n",
        "\n",
        "    @abstractmethod\n",
        "    def make_noise(self):\n",
        "        \"\"\" A function that prints a characteristic noise to terminal \"\"\"\n",
        "\n",
        "    @abstractmethod\n",
        "    def _set_mood(self):\n",
        "        \"\"\" A function that sets the mood of the animal \"\"\"\n",
        "\n",
        "    @property\n",
        "    def animal_species(self) -> str:\n",
        "        return self.species\n",
        "\n",
        "    @property\n",
        "    def animal_age(self) -> int:\n",
        "        return self.age\n",
        "\n",
        "    @property\n",
        "    def animal_noise(self) -> str:\n",
        "        return self.noise\n",
        "\n",
        "    @property\n",
        "    def animal_mood(self) -> str:\n",
        "        return self.mood\n",
        "\n",
        "    @property\n",
        "    def animal_name(self) -> str:\n",
        "        return self.name\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "81fPhWiicMvj"
      },
      "source": [
        "So above we have defined an \"abstract class\" for all our animals. The idea is that we can have other classes inherit from this base class; a `Cat` class, a `Dog`, etc. By setting `@abstractmethod` before `make_noise` and `_set_mood` we say that any class inheriting from this base class, must have a custom-made definition of these functions; but we leave it to the inheriting classes to choose exactly how these are defined.\n",
        "\n",
        "So the abstract class is abstract in the sense that it provides nothing more than structure and checks on the input. We use the built-in `isinstance()` which takes two arguments, an object and a class and returns `True` if the given class is anywhere in the inheritance chain of the object's class.\n",
        "\n",
        "Let's now define a `Cat` class that inherits from this abstract class."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 30,
      "metadata": {
        "id": "Efotb-LNSUiR"
      },
      "outputs": [],
      "source": [
        "import numpy as np\n",
        "class Cat(Animal):\n",
        "    \"\"\" A general cat class \"\"\"\n",
        "    def __init__(self, mood, age, name):\n",
        "        self.noise = 'meow'\n",
        "        self.species = 'cat'\n",
        "        super().__init__(noise = self.noise, species = self.species, mood = mood, age = age, name = name)\n",
        "\n",
        "    def make_noise(self):\n",
        "        print(f'{self.name} says: {self.noise}')\n",
        "        return\n",
        "\n",
        "    def _set_mood(self):\n",
        "        if self.mood == 'angry':\n",
        "            self.noise = self.noise.upper() + '!!'\n",
        "        else:\n",
        "            pass\n",
        "    @abstractmethod\n",
        "    def pet(self):\n",
        "        \"\"\" A function that defines how the cat responds to petting \"\"\"\n",
        ""
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "7mavGG7XdSfv"
      },
      "source": [
        "The Python super() function returns objects represented in the parent’s class and is very useful in  multiple and multilevel inheritances to find which class the child class is extending first. In other words, it allows us to pass on arguments from the `__init__` in `Cat` to the `__init__` function in `Animal`.\n",
        "\n",
        "Also; notice that we in the `Cat` class define the `make_noise` and `_set_mood` function, as required by our abstract base class `Animal`.\n",
        "\n",
        "In addition, we've hard-coded the `noise` and `species` member variables. We have also added yet another abstract method called `pet`. Here we effectivly say, that any cat will have to implement a cat-specific definition of how it responds to being pet.\n",
        "\n",
        "Let's now define a `Garfield` class that inherits from this `Cat` class!\n",
        "\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 31,
      "metadata": {
        "id": "hKf04axVWtn2"
      },
      "outputs": [],
      "source": [
        "class Garfield(Cat):\n",
        "    def __init__(self, mood):\n",
        "        self.age = 5\n",
        "        self.name = \"Garfield\"\n",
        "        self.mood = mood\n",
        "        super().__init__(self.mood, age = self.age, name = self.name)\n",
        "\n",
        "    def pet(self):\n",
        "        if np.random.randint(0,10) > 7:\n",
        "            print(f'{self.name} lets you continue petting.')\n",
        "        else:\n",
        "            print(f'{self.name} decides to go somewhere else. You love this disaffection, and {self.name} knows it.')\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "MqXdKl7KeYcc"
      },
      "source": [
        "Above we have hard-coded the age of Garfield and his name. The only argument to be specified now, is the mood of Garfield.\n",
        "\n",
        "\n",
        "Let's now go through similar steps for Snoopy!"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 32,
      "metadata": {
        "id": "qOCf9FpzUl7m"
      },
      "outputs": [],
      "source": [
        "class Dog(Animal):\n",
        "    \"\"\" A general dog class \"\"\"\n",
        "    def __init__(self, mood, age, name):\n",
        "        self.noise = 'wuff'\n",
        "        self.species = 'dog'\n",
        "        super().__init__(noise = self.noise, species = self.species, mood = mood, age = age, name = name)\n",
        "\n",
        "    def make_noise(self):\n",
        "        print(f'{self.name} says: {self.noise}')\n",
        "        return\n",
        "\n",
        "    def _set_mood(self):\n",
        "        if self.mood == 'angry':\n",
        "            self.noise = self.noise.upper() + '!!' +  self.noise.upper() + '!!'\n",
        "        else:\n",
        "            pass\n",
        "    @abstractmethod\n",
        "    def fetch(self, item: str):\n",
        "        \"\"\" A function that defines how the dog responds to being asked to fetch \"\"\""
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 33,
      "metadata": {
        "id": "W5meyg65X4lk"
      },
      "outputs": [],
      "source": [
        "class Snoopy(Dog):\n",
        "    def __init__(self, mood):\n",
        "        self.age = 7\n",
        "        self.name = \"Snoopy\"\n",
        "        self.mood = mood\n",
        "        super().__init__(mood = self.mood, age = self.age, name = self.name)\n",
        "\n",
        "    def fetch(self, item: str):\n",
        "        if item in ['stick', 'ball']:\n",
        "            print(f'{self.name} runs rapidly after the {item} and returns it proudly to you.')\n",
        "        else:\n",
        "            print(f'{self.name} is not interested in {item}.')"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "jTSyySzLeqUq"
      },
      "source": [
        "OK. Let's instantiate a Garfield and Snoopy and interact with them."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 34,
      "metadata": {
        "id": "yupaBhg3TRla"
      },
      "outputs": [],
      "source": [
        "garfield = Garfield(mood = 'happy')"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 35,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "dUJXTN-6Y9xJ",
        "outputId": "9f5a1f92-43ce-4e50-8bc6-2dc55db8fd18"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Garfield decides to go somewhere else. You love this disaffection, and Garfield knows it.\n"
          ]
        }
      ],
      "source": [
        "garfield.pet()"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 36,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "sVfd4D0zZlC8",
        "outputId": "0886a633-6011-4fe2-80e0-5f31a91d7c24"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Snoopy is not interested in good python pratices.\n"
          ]
        }
      ],
      "source": [
        "snoopy = Snoopy(mood = 'angry')\n",
        "snoopy.fetch('good python pratices')"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 37,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "EPJR5h5oQkdX",
        "outputId": "d79a6599-2d5b-40ef-de13-d3191c7354eb"
      },
      "outputs": [
        {
          "output_type": "stream",
          "name": "stdout",
          "text": [
            "Box contains: \n",
            "cats : \n",
            "        Garfield \n",
            "dogs : \n",
            "        Snoopy \n",
            "\n"
          ]
        }
      ],
      "source": [
        "my_box = BoxOfAnimals(animal = garfield)\n",
        "my_box.put(snoopy)\n",
        "\n",
        "my_box.print_inventory()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "2TTzO9xie0YG"
      },
      "source": [
        "OK, so to summarize:\n",
        "\n",
        "By increasing the level of abstracting we've been able to:\n",
        "\n",
        "\n",
        "\n",
        "1.   Place input checks in 1 single place (fewer lines of code)\n",
        "2.   Making a Garfield now only requires 1 argument; simpler to use\n",
        "3.   By creating seperate classes for Cats and Dogs, we've been able to implement species-specific behavior for each; how they make noise and how we interact with them (petting, fetching)\n",
        "4.   By creating different subclasses for cats and dogs, we can mimic personality by defining how a specific cat responds to interactions\n",
        "5.   The code is now structured in such a way that it is straight forward to create new types of animals and new versions of those that already exists. (If we had stayed with our initial choice of structure, adding another 10 different species would make that single class quite complicated.)\n",
        "\n",
        "\n",
        "\n"
      ]
    },
    {
      "cell_type": "code",
      "source": [],
      "metadata": {
        "id": "p4LuG0F8uQjO"
      },
      "execution_count": null,
      "outputs": []
    }
  ],
  "metadata": {
    "colab": {
      "provenance": [],
      "toc_visible": true
    },
    "kernelspec": {
      "display_name": "Python 3 (ipykernel)",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.11.1"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}